from __future__ import unicode_literals

from pybtex.bibtex.interpreter import (
    FunctionLiteral, Identifier, Integer, QuotedVar, String
)

bst = [[u'ENTRY',
  [Identifier(u'address'),
   Identifier(u'author'),
   Identifier(u'booktitle'),
   Identifier(u'chapter'),
   Identifier(u'edition'),
   Identifier(u'editor'),
   Identifier(u'howpublished'),
   Identifier(u'institution'),
   Identifier(u'journal'),
   Identifier(u'key'),
   Identifier(u'month'),
   Identifier(u'note'),
   Identifier(u'number'),
   Identifier(u'organization'),
   Identifier(u'pages'),
   Identifier(u'publisher'),
   Identifier(u'school'),
   Identifier(u'series'),
   Identifier(u'title'),
   Identifier(u'type'),
   Identifier(u'volume'),
   Identifier(u'year')],
  [],
  [Identifier(u'label')]],
 [u'INTEGERS',
  [Identifier(u'output.state'),
   Identifier(u'before.all'),
   Identifier(u'mid.sentence'),
   Identifier(u'after.sentence'),
   Identifier(u'after.block')]],
 [u'FUNCTION',
  [Identifier(u'init.state.consts')],
  [Integer(0),
   QuotedVar(u'before.all'),
   Identifier(u':='),
   Integer(1),
   QuotedVar(u'mid.sentence'),
   Identifier(u':='),
   Integer(2),
   QuotedVar(u'after.sentence'),
   Identifier(u':='),
   Integer(3),
   QuotedVar(u'after.block'),
   Identifier(u':=')]],
 [u'STRINGS', [Identifier(u's'), Identifier(u't')]],
 [u'FUNCTION',
  [Identifier(u'output.nonnull')],
  [QuotedVar(u's'),
   Identifier(u':='),
   Identifier(u'output.state'),
   Identifier(u'mid.sentence'),
   Identifier(u'='),
   FunctionLiteral([String(u', '), Identifier(u'*'), Identifier(u'write$')]),
   FunctionLiteral([Identifier(u'output.state'), Identifier(u'after.block'), Identifier(u'='), FunctionLiteral([Identifier(u'add.period$'), Identifier(u'write$'), Identifier(u'newline$'), String(u'\\newblock '), Identifier(u'write$')]), FunctionLiteral([Identifier(u'output.state'), Identifier(u'before.all'), Identifier(u'='), QuotedVar(u'write$'), FunctionLiteral([Identifier(u'add.period$'), String(u' '), Identifier(u'*'), Identifier(u'write$')]), Identifier(u'if$')]), Identifier(u'if$'), Identifier(u'mid.sentence'), QuotedVar(u'output.state'), Identifier(u':=')]),
   Identifier(u'if$'),
   Identifier(u's')]],
 [u'FUNCTION',
  [Identifier(u'output')],
  [Identifier(u'duplicate$'),
   Identifier(u'empty$'),
   QuotedVar(u'pop$'),
   QuotedVar(u'output.nonnull'),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'output.check')],
  [QuotedVar(u't'),
   Identifier(u':='),
   Identifier(u'duplicate$'),
   Identifier(u'empty$'),
   FunctionLiteral([Identifier(u'pop$'), String(u'empty '), Identifier(u't'), Identifier(u'*'), String(u' in '), Identifier(u'*'), Identifier(u'cite$'), Identifier(u'*'), Identifier(u'warning$')]),
   QuotedVar(u'output.nonnull'),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'output.bibitem')],
  [Identifier(u'newline$'),
   String(u'\\bibitem{'),
   Identifier(u'write$'),
   Identifier(u'cite$'),
   Identifier(u'write$'),
   String(u'}'),
   Identifier(u'write$'),
   Identifier(u'newline$'),
   String(u''),
   Identifier(u'before.all'),
   QuotedVar(u'output.state'),
   Identifier(u':=')]],
 [u'FUNCTION',
  [Identifier(u'fin.entry')],
  [Identifier(u'add.period$'),
   Identifier(u'write$'),
   Identifier(u'newline$')]],
 [u'FUNCTION',
  [Identifier(u'new.block')],
  [Identifier(u'output.state'),
   Identifier(u'before.all'),
   Identifier(u'='),
   QuotedVar(u'skip$'),
   FunctionLiteral([Identifier(u'after.block'), QuotedVar(u'output.state'), Identifier(u':=')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'new.sentence')],
  [Identifier(u'output.state'),
   Identifier(u'after.block'),
   Identifier(u'='),
   QuotedVar(u'skip$'),
   FunctionLiteral([Identifier(u'output.state'), Identifier(u'before.all'), Identifier(u'='), QuotedVar(u'skip$'), FunctionLiteral([Identifier(u'after.sentence'), QuotedVar(u'output.state'), Identifier(u':=')]), Identifier(u'if$')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'not')],
  [FunctionLiteral([Integer(0)]),
   FunctionLiteral([Integer(1)]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'and')],
  [QuotedVar(u'skip$'),
   FunctionLiteral([Identifier(u'pop$'), Integer(0)]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'or')],
  [FunctionLiteral([Identifier(u'pop$'), Integer(1)]),
   QuotedVar(u'skip$'),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'new.block.checka')],
  [Identifier(u'empty$'),
   QuotedVar(u'skip$'),
   QuotedVar(u'new.block'),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'new.block.checkb')],
  [Identifier(u'empty$'),
   Identifier(u'swap$'),
   Identifier(u'empty$'),
   Identifier(u'and'),
   QuotedVar(u'skip$'),
   QuotedVar(u'new.block'),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'new.sentence.checka')],
  [Identifier(u'empty$'),
   QuotedVar(u'skip$'),
   QuotedVar(u'new.sentence'),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'new.sentence.checkb')],
  [Identifier(u'empty$'),
   Identifier(u'swap$'),
   Identifier(u'empty$'),
   Identifier(u'and'),
   QuotedVar(u'skip$'),
   QuotedVar(u'new.sentence'),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'field.or.null')],
  [Identifier(u'duplicate$'),
   Identifier(u'empty$'),
   FunctionLiteral([Identifier(u'pop$'), String(u'')]),
   QuotedVar(u'skip$'),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'emphasize')],
  [Identifier(u'duplicate$'),
   Identifier(u'empty$'),
   FunctionLiteral([Identifier(u'pop$'), String(u'')]),
   FunctionLiteral([String(u'{\\em '), Identifier(u'swap$'), Identifier(u'*'), String(u'}'), Identifier(u'*')]),
   Identifier(u'if$')]],
 [u'INTEGERS',
  [Identifier(u'nameptr'), Identifier(u'namesleft'), Identifier(u'numnames')]],
 [u'FUNCTION',
  [Identifier(u'format.names')],
  [QuotedVar(u's'),
   Identifier(u':='),
   Integer(1),
   QuotedVar(u'nameptr'),
   Identifier(u':='),
   Identifier(u's'),
   Identifier(u'num.names$'),
   QuotedVar(u'numnames'),
   Identifier(u':='),
   Identifier(u'numnames'),
   QuotedVar(u'namesleft'),
   Identifier(u':='),
   FunctionLiteral([Identifier(u'namesleft'), Integer(0), Identifier(u'>')]),
   FunctionLiteral([Identifier(u's'), Identifier(u'nameptr'), String(u'{ff~}{vv~}{ll}{, jj}'), Identifier(u'format.name$'), QuotedVar(u't'), Identifier(u':='), Identifier(u'nameptr'), Integer(1), Identifier(u'>'), FunctionLiteral([Identifier(u'namesleft'), Integer(1), Identifier(u'>'), FunctionLiteral([String(u', '), Identifier(u'*'), Identifier(u't'), Identifier(u'*')]), FunctionLiteral([Identifier(u'numnames'), Integer(2), Identifier(u'>'), FunctionLiteral([String(u','), Identifier(u'*')]), QuotedVar(u'skip$'), Identifier(u'if$'), Identifier(u't'), String(u'others'), Identifier(u'='), FunctionLiteral([String(u' et~al.'), Identifier(u'*')]), FunctionLiteral([String(u' and '), Identifier(u'*'), Identifier(u't'), Identifier(u'*')]), Identifier(u'if$')]), Identifier(u'if$')]), QuotedVar(u't'), Identifier(u'if$'), Identifier(u'nameptr'), Integer(1), Identifier(u'+'), QuotedVar(u'nameptr'), Identifier(u':='), Identifier(u'namesleft'), Integer(1), Identifier(u'-'), QuotedVar(u'namesleft'), Identifier(u':=')]),
   Identifier(u'while$')]],
 [u'FUNCTION',
  [Identifier(u'format.authors')],
  [Identifier(u'author'),
   Identifier(u'empty$'),
   FunctionLiteral([String(u'')]),
   FunctionLiteral([Identifier(u'author'), Identifier(u'format.names')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'format.editors')],
  [Identifier(u'editor'),
   Identifier(u'empty$'),
   FunctionLiteral([String(u'')]),
   FunctionLiteral([Identifier(u'editor'), Identifier(u'format.names'), Identifier(u'editor'), Identifier(u'num.names$'), Integer(1), Identifier(u'>'), FunctionLiteral([String(u', editors'), Identifier(u'*')]), FunctionLiteral([String(u', editor'), Identifier(u'*')]), Identifier(u'if$')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'format.title')],
  [Identifier(u'title'),
   Identifier(u'empty$'),
   FunctionLiteral([String(u'')]),
   FunctionLiteral([Identifier(u'title'), String(u't'), Identifier(u'change.case$')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'n.dashify')],
  [QuotedVar(u't'),
   Identifier(u':='),
   String(u''),
   FunctionLiteral([Identifier(u't'), Identifier(u'empty$'), Identifier(u'not')]),
   FunctionLiteral([Identifier(u't'), Integer(1), Integer(1), Identifier(u'substring$'), String(u'-'), Identifier(u'='), FunctionLiteral([Identifier(u't'), Integer(1), Integer(2), Identifier(u'substring$'), String(u'--'), Identifier(u'='), Identifier(u'not'), FunctionLiteral([String(u'--'), Identifier(u'*'), Identifier(u't'), Integer(2), Identifier(u'global.max$'), Identifier(u'substring$'), QuotedVar(u't'), Identifier(u':=')]), FunctionLiteral([FunctionLiteral([Identifier(u't'), Integer(1), Integer(1), Identifier(u'substring$'), String(u'-'), Identifier(u'=')]), FunctionLiteral([String(u'-'), Identifier(u'*'), Identifier(u't'), Integer(2), Identifier(u'global.max$'), Identifier(u'substring$'), QuotedVar(u't'), Identifier(u':=')]), Identifier(u'while$')]), Identifier(u'if$')]), FunctionLiteral([Identifier(u't'), Integer(1), Integer(1), Identifier(u'substring$'), Identifier(u'*'), Identifier(u't'), Integer(2), Identifier(u'global.max$'), Identifier(u'substring$'), QuotedVar(u't'), Identifier(u':=')]), Identifier(u'if$')]),
   Identifier(u'while$')]],
 [u'FUNCTION',
  [Identifier(u'format.date')],
  [Identifier(u'year'),
   Identifier(u'empty$'),
   FunctionLiteral([Identifier(u'month'), Identifier(u'empty$'), FunctionLiteral([String(u'')]), FunctionLiteral([String(u"there's a month but no year in "), Identifier(u'cite$'), Identifier(u'*'), Identifier(u'warning$'), Identifier(u'month')]), Identifier(u'if$')]),
   FunctionLiteral([Identifier(u'month'), Identifier(u'empty$'), QuotedVar(u'year'), FunctionLiteral([Identifier(u'month'), String(u' '), Identifier(u'*'), Identifier(u'year'), Identifier(u'*')]), Identifier(u'if$')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'format.btitle')],
  [Identifier(u'title'), Identifier(u'emphasize')]],
 [u'FUNCTION',
  [Identifier(u'tie.or.space.connect')],
  [Identifier(u'duplicate$'),
   Identifier(u'text.length$'),
   Integer(3),
   Identifier(u'<'),
   FunctionLiteral([String(u'~')]),
   FunctionLiteral([String(u' ')]),
   Identifier(u'if$'),
   Identifier(u'swap$'),
   Identifier(u'*'),
   Identifier(u'*')]],
 [u'FUNCTION',
  [Identifier(u'either.or.check')],
  [Identifier(u'empty$'),
   QuotedVar(u'pop$'),
   FunctionLiteral([String(u"can't use both "), Identifier(u'swap$'), Identifier(u'*'), String(u' fields in '), Identifier(u'*'), Identifier(u'cite$'), Identifier(u'*'), Identifier(u'warning$')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'format.bvolume')],
  [Identifier(u'volume'),
   Identifier(u'empty$'),
   FunctionLiteral([String(u'')]),
   FunctionLiteral([String(u'volume'), Identifier(u'volume'), Identifier(u'tie.or.space.connect'), Identifier(u'series'), Identifier(u'empty$'), QuotedVar(u'skip$'), FunctionLiteral([String(u' of '), Identifier(u'*'), Identifier(u'series'), Identifier(u'emphasize'), Identifier(u'*')]), Identifier(u'if$'), String(u'volume and number'), Identifier(u'number'), Identifier(u'either.or.check')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'format.number.series')],
  [Identifier(u'volume'),
   Identifier(u'empty$'),
   FunctionLiteral([Identifier(u'number'), Identifier(u'empty$'), FunctionLiteral([Identifier(u'series'), Identifier(u'field.or.null')]), FunctionLiteral([Identifier(u'output.state'), Identifier(u'mid.sentence'), Identifier(u'='), FunctionLiteral([String(u'number')]), FunctionLiteral([String(u'Number')]), Identifier(u'if$'), Identifier(u'number'), Identifier(u'tie.or.space.connect'), Identifier(u'series'), Identifier(u'empty$'), FunctionLiteral([String(u"there's a number but no series in "), Identifier(u'cite$'), Identifier(u'*'), Identifier(u'warning$')]), FunctionLiteral([String(u' in '), Identifier(u'*'), Identifier(u'series'), Identifier(u'*')]), Identifier(u'if$')]), Identifier(u'if$')]),
   FunctionLiteral([String(u'')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'format.edition')],
  [Identifier(u'edition'),
   Identifier(u'empty$'),
   FunctionLiteral([String(u'')]),
   FunctionLiteral([Identifier(u'output.state'), Identifier(u'mid.sentence'), Identifier(u'='), FunctionLiteral([Identifier(u'edition'), String(u'l'), Identifier(u'change.case$'), String(u' edition'), Identifier(u'*')]), FunctionLiteral([Identifier(u'edition'), String(u't'), Identifier(u'change.case$'), String(u' edition'), Identifier(u'*')]), Identifier(u'if$')]),
   Identifier(u'if$')]],
 [u'INTEGERS', [Identifier(u'multiresult')]],
 [u'FUNCTION',
  [Identifier(u'multi.page.check')],
  [QuotedVar(u't'),
   Identifier(u':='),
   Integer(0),
   QuotedVar(u'multiresult'),
   Identifier(u':='),
   FunctionLiteral([Identifier(u'multiresult'), Identifier(u'not'), Identifier(u't'), Identifier(u'empty$'), Identifier(u'not'), Identifier(u'and')]),
   FunctionLiteral([Identifier(u't'), Integer(1), Integer(1), Identifier(u'substring$'), Identifier(u'duplicate$'), String(u'-'), Identifier(u'='), Identifier(u'swap$'), Identifier(u'duplicate$'), String(u','), Identifier(u'='), Identifier(u'swap$'), String(u'+'), Identifier(u'='), Identifier(u'or'), Identifier(u'or'), FunctionLiteral([Integer(1), QuotedVar(u'multiresult'), Identifier(u':=')]), FunctionLiteral([Identifier(u't'), Integer(2), Identifier(u'global.max$'), Identifier(u'substring$'), QuotedVar(u't'), Identifier(u':=')]), Identifier(u'if$')]),
   Identifier(u'while$'),
   Identifier(u'multiresult')]],
 [u'FUNCTION',
  [Identifier(u'format.pages')],
  [Identifier(u'pages'),
   Identifier(u'empty$'),
   FunctionLiteral([String(u'')]),
   FunctionLiteral([Identifier(u'pages'), Identifier(u'multi.page.check'), FunctionLiteral([String(u'pages'), Identifier(u'pages'), Identifier(u'n.dashify'), Identifier(u'tie.or.space.connect')]), FunctionLiteral([String(u'page'), Identifier(u'pages'), Identifier(u'tie.or.space.connect')]), Identifier(u'if$')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'format.vol.num.pages')],
  [Identifier(u'volume'),
   Identifier(u'field.or.null'),
   Identifier(u'number'),
   Identifier(u'empty$'),
   QuotedVar(u'skip$'),
   FunctionLiteral([String(u'('), Identifier(u'number'), Identifier(u'*'), String(u')'), Identifier(u'*'), Identifier(u'*'), Identifier(u'volume'), Identifier(u'empty$'), FunctionLiteral([String(u"there's a number but no volume in "), Identifier(u'cite$'), Identifier(u'*'), Identifier(u'warning$')]), QuotedVar(u'skip$'), Identifier(u'if$')]),
   Identifier(u'if$'),
   Identifier(u'pages'),
   Identifier(u'empty$'),
   QuotedVar(u'skip$'),
   FunctionLiteral([Identifier(u'duplicate$'), Identifier(u'empty$'), FunctionLiteral([Identifier(u'pop$'), Identifier(u'format.pages')]), FunctionLiteral([String(u':'), Identifier(u'*'), Identifier(u'pages'), Identifier(u'n.dashify'), Identifier(u'*')]), Identifier(u'if$')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'format.chapter.pages')],
  [Identifier(u'chapter'),
   Identifier(u'empty$'),
   QuotedVar(u'format.pages'),
   FunctionLiteral([Identifier(u'type'), Identifier(u'empty$'), FunctionLiteral([String(u'chapter')]), FunctionLiteral([Identifier(u'type'), String(u'l'), Identifier(u'change.case$')]), Identifier(u'if$'), Identifier(u'chapter'), Identifier(u'tie.or.space.connect'), Identifier(u'pages'), Identifier(u'empty$'), QuotedVar(u'skip$'), FunctionLiteral([String(u', '), Identifier(u'*'), Identifier(u'format.pages'), Identifier(u'*')]), Identifier(u'if$')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'format.in.ed.booktitle')],
  [Identifier(u'booktitle'),
   Identifier(u'empty$'),
   FunctionLiteral([String(u'')]),
   FunctionLiteral([Identifier(u'editor'), Identifier(u'empty$'), FunctionLiteral([String(u'In '), Identifier(u'booktitle'), Identifier(u'emphasize'), Identifier(u'*')]), FunctionLiteral([String(u'In '), Identifier(u'format.editors'), Identifier(u'*'), String(u', '), Identifier(u'*'), Identifier(u'booktitle'), Identifier(u'emphasize'), Identifier(u'*')]), Identifier(u'if$')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'empty.misc.check')],
  [Identifier(u'author'),
   Identifier(u'empty$'),
   Identifier(u'title'),
   Identifier(u'empty$'),
   Identifier(u'howpublished'),
   Identifier(u'empty$'),
   Identifier(u'month'),
   Identifier(u'empty$'),
   Identifier(u'year'),
   Identifier(u'empty$'),
   Identifier(u'note'),
   Identifier(u'empty$'),
   Identifier(u'and'),
   Identifier(u'and'),
   Identifier(u'and'),
   Identifier(u'and'),
   Identifier(u'and'),
   Identifier(u'key'),
   Identifier(u'empty$'),
   Identifier(u'not'),
   Identifier(u'and'),
   FunctionLiteral([String(u'all relevant fields are empty in '), Identifier(u'cite$'), Identifier(u'*'), Identifier(u'warning$')]),
   QuotedVar(u'skip$'),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'format.thesis.type')],
  [Identifier(u'type'),
   Identifier(u'empty$'),
   QuotedVar(u'skip$'),
   FunctionLiteral([Identifier(u'pop$'), Identifier(u'type'), String(u't'), Identifier(u'change.case$')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'format.tr.number')],
  [Identifier(u'type'),
   Identifier(u'empty$'),
   FunctionLiteral([String(u'Technical Report')]),
   QuotedVar(u'type'),
   Identifier(u'if$'),
   Identifier(u'number'),
   Identifier(u'empty$'),
   FunctionLiteral([String(u't'), Identifier(u'change.case$')]),
   FunctionLiteral([Identifier(u'number'), Identifier(u'tie.or.space.connect')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'format.article.crossref')],
  [Identifier(u'key'),
   Identifier(u'empty$'),
   FunctionLiteral([Identifier(u'journal'), Identifier(u'empty$'), FunctionLiteral([String(u'need key or journal for '), Identifier(u'cite$'), Identifier(u'*'), String(u' to crossref '), Identifier(u'*'), Identifier(u'crossref'), Identifier(u'*'), Identifier(u'warning$'), String(u'')]), FunctionLiteral([String(u'In {\\em '), Identifier(u'journal'), Identifier(u'*'), String(u'\\/}'), Identifier(u'*')]), Identifier(u'if$')]),
   FunctionLiteral([String(u'In '), Identifier(u'key'), Identifier(u'*')]),
   Identifier(u'if$'),
   String(u' \\cite{'),
   Identifier(u'*'),
   Identifier(u'crossref'),
   Identifier(u'*'),
   String(u'}'),
   Identifier(u'*')]],
 [u'FUNCTION',
  [Identifier(u'format.crossref.editor')],
  [Identifier(u'editor'),
   Integer(1),
   String(u'{vv~}{ll}'),
   Identifier(u'format.name$'),
   Identifier(u'editor'),
   Identifier(u'num.names$'),
   Identifier(u'duplicate$'),
   Integer(2),
   Identifier(u'>'),
   FunctionLiteral([Identifier(u'pop$'), String(u' et~al.'), Identifier(u'*')]),
   FunctionLiteral([Integer(2), Identifier(u'<'), QuotedVar(u'skip$'), FunctionLiteral([Identifier(u'editor'), Integer(2), String(u'{ff }{vv }{ll}{ jj}'), Identifier(u'format.name$'), String(u'others'), Identifier(u'='), FunctionLiteral([String(u' et~al.'), Identifier(u'*')]), FunctionLiteral([String(u' and '), Identifier(u'*'), Identifier(u'editor'), Integer(2), String(u'{vv~}{ll}'), Identifier(u'format.name$'), Identifier(u'*')]), Identifier(u'if$')]), Identifier(u'if$')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'format.book.crossref')],
  [Identifier(u'volume'),
   Identifier(u'empty$'),
   FunctionLiteral([String(u'empty volume in '), Identifier(u'cite$'), Identifier(u'*'), String(u"'s crossref of "), Identifier(u'*'), Identifier(u'crossref'), Identifier(u'*'), Identifier(u'warning$'), String(u'In ')]),
   FunctionLiteral([String(u'Volume'), Identifier(u'volume'), Identifier(u'tie.or.space.connect'), String(u' of '), Identifier(u'*')]),
   Identifier(u'if$'),
   Identifier(u'editor'),
   Identifier(u'empty$'),
   Identifier(u'editor'),
   Identifier(u'field.or.null'),
   Identifier(u'author'),
   Identifier(u'field.or.null'),
   Identifier(u'='),
   Identifier(u'or'),
   FunctionLiteral([Identifier(u'key'), Identifier(u'empty$'), FunctionLiteral([Identifier(u'series'), Identifier(u'empty$'), FunctionLiteral([String(u'need editor, key, or series for '), Identifier(u'cite$'), Identifier(u'*'), String(u' to crossref '), Identifier(u'*'), Identifier(u'crossref'), Identifier(u'*'), Identifier(u'warning$'), String(u''), Identifier(u'*')]), FunctionLiteral([String(u'{\\em '), Identifier(u'*'), Identifier(u'series'), Identifier(u'*'), String(u'\\/}'), Identifier(u'*')]), Identifier(u'if$')]), FunctionLiteral([Identifier(u'key'), Identifier(u'*')]), Identifier(u'if$')]),
   FunctionLiteral([Identifier(u'format.crossref.editor'), Identifier(u'*')]),
   Identifier(u'if$'),
   String(u' \\cite{'),
   Identifier(u'*'),
   Identifier(u'crossref'),
   Identifier(u'*'),
   String(u'}'),
   Identifier(u'*')]],
 [u'FUNCTION',
  [Identifier(u'format.incoll.inproc.crossref')],
  [Identifier(u'editor'),
   Identifier(u'empty$'),
   Identifier(u'editor'),
   Identifier(u'field.or.null'),
   Identifier(u'author'),
   Identifier(u'field.or.null'),
   Identifier(u'='),
   Identifier(u'or'),
   FunctionLiteral([Identifier(u'key'), Identifier(u'empty$'), FunctionLiteral([Identifier(u'booktitle'), Identifier(u'empty$'), FunctionLiteral([String(u'need editor, key, or booktitle for '), Identifier(u'cite$'), Identifier(u'*'), String(u' to crossref '), Identifier(u'*'), Identifier(u'crossref'), Identifier(u'*'), Identifier(u'warning$'), String(u'')]), FunctionLiteral([String(u'In {\\em '), Identifier(u'booktitle'), Identifier(u'*'), String(u'\\/}'), Identifier(u'*')]), Identifier(u'if$')]), FunctionLiteral([String(u'In '), Identifier(u'key'), Identifier(u'*')]), Identifier(u'if$')]),
   FunctionLiteral([String(u'In '), Identifier(u'format.crossref.editor'), Identifier(u'*')]),
   Identifier(u'if$'),
   String(u' \\cite{'),
   Identifier(u'*'),
   Identifier(u'crossref'),
   Identifier(u'*'),
   String(u'}'),
   Identifier(u'*')]],
 [u'FUNCTION',
  [Identifier(u'article')],
  [Identifier(u'output.bibitem'),
   Identifier(u'format.authors'),
   String(u'author'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'format.title'),
   String(u'title'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'crossref'),
   Identifier(u'missing$'),
   FunctionLiteral([Identifier(u'journal'), Identifier(u'emphasize'), String(u'journal'), Identifier(u'output.check'), Identifier(u'format.vol.num.pages'), Identifier(u'output'), Identifier(u'format.date'), String(u'year'), Identifier(u'output.check')]),
   FunctionLiteral([Identifier(u'format.article.crossref'), Identifier(u'output.nonnull'), Identifier(u'format.pages'), Identifier(u'output')]),
   Identifier(u'if$'),
   Identifier(u'new.block'),
   Identifier(u'note'),
   Identifier(u'output'),
   Identifier(u'fin.entry')]],
 [u'FUNCTION',
  [Identifier(u'book')],
  [Identifier(u'output.bibitem'),
   Identifier(u'author'),
   Identifier(u'empty$'),
   FunctionLiteral([Identifier(u'format.editors'), String(u'author and editor'), Identifier(u'output.check')]),
   FunctionLiteral([Identifier(u'format.authors'), Identifier(u'output.nonnull'), Identifier(u'crossref'), Identifier(u'missing$'), FunctionLiteral([String(u'author and editor'), Identifier(u'editor'), Identifier(u'either.or.check')]), QuotedVar(u'skip$'), Identifier(u'if$')]),
   Identifier(u'if$'),
   Identifier(u'new.block'),
   Identifier(u'format.btitle'),
   String(u'title'),
   Identifier(u'output.check'),
   Identifier(u'crossref'),
   Identifier(u'missing$'),
   FunctionLiteral([Identifier(u'format.bvolume'), Identifier(u'output'), Identifier(u'new.block'), Identifier(u'format.number.series'), Identifier(u'output'), Identifier(u'new.sentence'), Identifier(u'publisher'), String(u'publisher'), Identifier(u'output.check'), Identifier(u'address'), Identifier(u'output')]),
   FunctionLiteral([Identifier(u'new.block'), Identifier(u'format.book.crossref'), Identifier(u'output.nonnull')]),
   Identifier(u'if$'),
   Identifier(u'format.edition'),
   Identifier(u'output'),
   Identifier(u'format.date'),
   String(u'year'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'note'),
   Identifier(u'output'),
   Identifier(u'fin.entry')]],
 [u'FUNCTION',
  [Identifier(u'booklet')],
  [Identifier(u'output.bibitem'),
   Identifier(u'format.authors'),
   Identifier(u'output'),
   Identifier(u'new.block'),
   Identifier(u'format.title'),
   String(u'title'),
   Identifier(u'output.check'),
   Identifier(u'howpublished'),
   Identifier(u'address'),
   Identifier(u'new.block.checkb'),
   Identifier(u'howpublished'),
   Identifier(u'output'),
   Identifier(u'address'),
   Identifier(u'output'),
   Identifier(u'format.date'),
   Identifier(u'output'),
   Identifier(u'new.block'),
   Identifier(u'note'),
   Identifier(u'output'),
   Identifier(u'fin.entry')]],
 [u'FUNCTION',
  [Identifier(u'inbook')],
  [Identifier(u'output.bibitem'),
   Identifier(u'author'),
   Identifier(u'empty$'),
   FunctionLiteral([Identifier(u'format.editors'), String(u'author and editor'), Identifier(u'output.check')]),
   FunctionLiteral([Identifier(u'format.authors'), Identifier(u'output.nonnull'), Identifier(u'crossref'), Identifier(u'missing$'), FunctionLiteral([String(u'author and editor'), Identifier(u'editor'), Identifier(u'either.or.check')]), QuotedVar(u'skip$'), Identifier(u'if$')]),
   Identifier(u'if$'),
   Identifier(u'new.block'),
   Identifier(u'format.btitle'),
   String(u'title'),
   Identifier(u'output.check'),
   Identifier(u'crossref'),
   Identifier(u'missing$'),
   FunctionLiteral([Identifier(u'format.bvolume'), Identifier(u'output'), Identifier(u'format.chapter.pages'), String(u'chapter and pages'), Identifier(u'output.check'), Identifier(u'new.block'), Identifier(u'format.number.series'), Identifier(u'output'), Identifier(u'new.sentence'), Identifier(u'publisher'), String(u'publisher'), Identifier(u'output.check'), Identifier(u'address'), Identifier(u'output')]),
   FunctionLiteral([Identifier(u'format.chapter.pages'), String(u'chapter and pages'), Identifier(u'output.check'), Identifier(u'new.block'), Identifier(u'format.book.crossref'), Identifier(u'output.nonnull')]),
   Identifier(u'if$'),
   Identifier(u'format.edition'),
   Identifier(u'output'),
   Identifier(u'format.date'),
   String(u'year'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'note'),
   Identifier(u'output'),
   Identifier(u'fin.entry')]],
 [u'FUNCTION',
  [Identifier(u'incollection')],
  [Identifier(u'output.bibitem'),
   Identifier(u'format.authors'),
   String(u'author'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'format.title'),
   String(u'title'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'crossref'),
   Identifier(u'missing$'),
   FunctionLiteral([Identifier(u'format.in.ed.booktitle'), String(u'booktitle'), Identifier(u'output.check'), Identifier(u'format.bvolume'), Identifier(u'output'), Identifier(u'format.number.series'), Identifier(u'output'), Identifier(u'format.chapter.pages'), Identifier(u'output'), Identifier(u'new.sentence'), Identifier(u'publisher'), String(u'publisher'), Identifier(u'output.check'), Identifier(u'address'), Identifier(u'output'), Identifier(u'format.edition'), Identifier(u'output'), Identifier(u'format.date'), String(u'year'), Identifier(u'output.check')]),
   FunctionLiteral([Identifier(u'format.incoll.inproc.crossref'), Identifier(u'output.nonnull'), Identifier(u'format.chapter.pages'), Identifier(u'output')]),
   Identifier(u'if$'),
   Identifier(u'new.block'),
   Identifier(u'note'),
   Identifier(u'output'),
   Identifier(u'fin.entry')]],
 [u'FUNCTION',
  [Identifier(u'inproceedings')],
  [Identifier(u'output.bibitem'),
   Identifier(u'format.authors'),
   String(u'author'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'format.title'),
   String(u'title'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'crossref'),
   Identifier(u'missing$'),
   FunctionLiteral([Identifier(u'format.in.ed.booktitle'), String(u'booktitle'), Identifier(u'output.check'), Identifier(u'format.bvolume'), Identifier(u'output'), Identifier(u'format.number.series'), Identifier(u'output'), Identifier(u'format.pages'), Identifier(u'output'), Identifier(u'address'), Identifier(u'empty$'), FunctionLiteral([Identifier(u'organization'), Identifier(u'publisher'), Identifier(u'new.sentence.checkb'), Identifier(u'organization'), Identifier(u'output'), Identifier(u'publisher'), Identifier(u'output'), Identifier(u'format.date'), String(u'year'), Identifier(u'output.check')]), FunctionLiteral([Identifier(u'address'), Identifier(u'output.nonnull'), Identifier(u'format.date'), String(u'year'), Identifier(u'output.check'), Identifier(u'new.sentence'), Identifier(u'organization'), Identifier(u'output'), Identifier(u'publisher'), Identifier(u'output')]), Identifier(u'if$')]),
   FunctionLiteral([Identifier(u'format.incoll.inproc.crossref'), Identifier(u'output.nonnull'), Identifier(u'format.pages'), Identifier(u'output')]),
   Identifier(u'if$'),
   Identifier(u'new.block'),
   Identifier(u'note'),
   Identifier(u'output'),
   Identifier(u'fin.entry')]],
 [u'FUNCTION', [Identifier(u'conference')], [Identifier(u'inproceedings')]],
 [u'FUNCTION',
  [Identifier(u'manual')],
  [Identifier(u'output.bibitem'),
   Identifier(u'author'),
   Identifier(u'empty$'),
   FunctionLiteral([Identifier(u'organization'), Identifier(u'empty$'), QuotedVar(u'skip$'), FunctionLiteral([Identifier(u'organization'), Identifier(u'output.nonnull'), Identifier(u'address'), Identifier(u'output')]), Identifier(u'if$')]),
   FunctionLiteral([Identifier(u'format.authors'), Identifier(u'output.nonnull')]),
   Identifier(u'if$'),
   Identifier(u'new.block'),
   Identifier(u'format.btitle'),
   String(u'title'),
   Identifier(u'output.check'),
   Identifier(u'author'),
   Identifier(u'empty$'),
   FunctionLiteral([Identifier(u'organization'), Identifier(u'empty$'), FunctionLiteral([Identifier(u'address'), Identifier(u'new.block.checka'), Identifier(u'address'), Identifier(u'output')]), QuotedVar(u'skip$'), Identifier(u'if$')]),
   FunctionLiteral([Identifier(u'organization'), Identifier(u'address'), Identifier(u'new.block.checkb'), Identifier(u'organization'), Identifier(u'output'), Identifier(u'address'), Identifier(u'output')]),
   Identifier(u'if$'),
   Identifier(u'format.edition'),
   Identifier(u'output'),
   Identifier(u'format.date'),
   Identifier(u'output'),
   Identifier(u'new.block'),
   Identifier(u'note'),
   Identifier(u'output'),
   Identifier(u'fin.entry')]],
 [u'FUNCTION',
  [Identifier(u'mastersthesis')],
  [Identifier(u'output.bibitem'),
   Identifier(u'format.authors'),
   String(u'author'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'format.title'),
   String(u'title'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   String(u"Master's thesis"),
   Identifier(u'format.thesis.type'),
   Identifier(u'output.nonnull'),
   Identifier(u'school'),
   String(u'school'),
   Identifier(u'output.check'),
   Identifier(u'address'),
   Identifier(u'output'),
   Identifier(u'format.date'),
   String(u'year'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'note'),
   Identifier(u'output'),
   Identifier(u'fin.entry')]],
 [u'FUNCTION',
  [Identifier(u'misc')],
  [Identifier(u'output.bibitem'),
   Identifier(u'format.authors'),
   Identifier(u'output'),
   Identifier(u'title'),
   Identifier(u'howpublished'),
   Identifier(u'new.block.checkb'),
   Identifier(u'format.title'),
   Identifier(u'output'),
   Identifier(u'howpublished'),
   Identifier(u'new.block.checka'),
   Identifier(u'howpublished'),
   Identifier(u'output'),
   Identifier(u'format.date'),
   Identifier(u'output'),
   Identifier(u'new.block'),
   Identifier(u'note'),
   Identifier(u'output'),
   Identifier(u'fin.entry'),
   Identifier(u'empty.misc.check')]],
 [u'FUNCTION',
  [Identifier(u'phdthesis')],
  [Identifier(u'output.bibitem'),
   Identifier(u'format.authors'),
   String(u'author'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'format.btitle'),
   String(u'title'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   String(u'PhD thesis'),
   Identifier(u'format.thesis.type'),
   Identifier(u'output.nonnull'),
   Identifier(u'school'),
   String(u'school'),
   Identifier(u'output.check'),
   Identifier(u'address'),
   Identifier(u'output'),
   Identifier(u'format.date'),
   String(u'year'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'note'),
   Identifier(u'output'),
   Identifier(u'fin.entry')]],
 [u'FUNCTION',
  [Identifier(u'proceedings')],
  [Identifier(u'output.bibitem'),
   Identifier(u'editor'),
   Identifier(u'empty$'),
   FunctionLiteral([Identifier(u'organization'), Identifier(u'output')]),
   FunctionLiteral([Identifier(u'format.editors'), Identifier(u'output.nonnull')]),
   Identifier(u'if$'),
   Identifier(u'new.block'),
   Identifier(u'format.btitle'),
   String(u'title'),
   Identifier(u'output.check'),
   Identifier(u'format.bvolume'),
   Identifier(u'output'),
   Identifier(u'format.number.series'),
   Identifier(u'output'),
   Identifier(u'address'),
   Identifier(u'empty$'),
   FunctionLiteral([Identifier(u'editor'), Identifier(u'empty$'), FunctionLiteral([Identifier(u'publisher'), Identifier(u'new.sentence.checka')]), FunctionLiteral([Identifier(u'organization'), Identifier(u'publisher'), Identifier(u'new.sentence.checkb'), Identifier(u'organization'), Identifier(u'output')]), Identifier(u'if$'), Identifier(u'publisher'), Identifier(u'output'), Identifier(u'format.date'), String(u'year'), Identifier(u'output.check')]),
   FunctionLiteral([Identifier(u'address'), Identifier(u'output.nonnull'), Identifier(u'format.date'), String(u'year'), Identifier(u'output.check'), Identifier(u'new.sentence'), Identifier(u'editor'), Identifier(u'empty$'), QuotedVar(u'skip$'), FunctionLiteral([Identifier(u'organization'), Identifier(u'output')]), Identifier(u'if$'), Identifier(u'publisher'), Identifier(u'output')]),
   Identifier(u'if$'),
   Identifier(u'new.block'),
   Identifier(u'note'),
   Identifier(u'output'),
   Identifier(u'fin.entry')]],
 [u'FUNCTION',
  [Identifier(u'techreport')],
  [Identifier(u'output.bibitem'),
   Identifier(u'format.authors'),
   String(u'author'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'format.title'),
   String(u'title'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'format.tr.number'),
   Identifier(u'output.nonnull'),
   Identifier(u'institution'),
   String(u'institution'),
   Identifier(u'output.check'),
   Identifier(u'address'),
   Identifier(u'output'),
   Identifier(u'format.date'),
   String(u'year'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'note'),
   Identifier(u'output'),
   Identifier(u'fin.entry')]],
 [u'FUNCTION',
  [Identifier(u'unpublished')],
  [Identifier(u'output.bibitem'),
   Identifier(u'format.authors'),
   String(u'author'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'format.title'),
   String(u'title'),
   Identifier(u'output.check'),
   Identifier(u'new.block'),
   Identifier(u'note'),
   String(u'note'),
   Identifier(u'output.check'),
   Identifier(u'format.date'),
   Identifier(u'output'),
   Identifier(u'fin.entry')]],
 [u'FUNCTION', [Identifier(u'default.type')], [Identifier(u'misc')]],
 [u'MACRO', [Identifier(u'jan')], [String(u'January')]],
 [u'MACRO', [Identifier(u'feb')], [String(u'February')]],
 [u'MACRO', [Identifier(u'mar')], [String(u'March')]],
 [u'MACRO', [Identifier(u'apr')], [String(u'April')]],
 [u'MACRO', [Identifier(u'may')], [String(u'May')]],
 [u'MACRO', [Identifier(u'jun')], [String(u'June')]],
 [u'MACRO', [Identifier(u'jul')], [String(u'July')]],
 [u'MACRO', [Identifier(u'aug')], [String(u'August')]],
 [u'MACRO', [Identifier(u'sep')], [String(u'September')]],
 [u'MACRO', [Identifier(u'oct')], [String(u'October')]],
 [u'MACRO', [Identifier(u'nov')], [String(u'November')]],
 [u'MACRO', [Identifier(u'dec')], [String(u'December')]],
 [u'MACRO', [Identifier(u'acmcs')], [String(u'ACM Computing Surveys')]],
 [u'MACRO', [Identifier(u'acta')], [String(u'Acta Informatica')]],
 [u'MACRO', [Identifier(u'cacm')], [String(u'Communications of the ACM')]],
 [u'MACRO',
  [Identifier(u'ibmjrd')],
  [String(u'IBM Journal of Research and Development')]],
 [u'MACRO', [Identifier(u'ibmsj')], [String(u'IBM Systems Journal')]],
 [u'MACRO',
  [Identifier(u'ieeese')],
  [String(u'IEEE Transactions on Software Engineering')]],
 [u'MACRO',
  [Identifier(u'ieeetc')],
  [String(u'IEEE Transactions on Computers')]],
 [u'MACRO',
  [Identifier(u'ieeetcad')],
  [String(u'IEEE Transactions on Computer-Aided Design of Integrated Circuits')]],
 [u'MACRO', [Identifier(u'ipl')], [String(u'Information Processing Letters')]],
 [u'MACRO', [Identifier(u'jacm')], [String(u'Journal of the ACM')]],
 [u'MACRO',
  [Identifier(u'jcss')],
  [String(u'Journal of Computer and System Sciences')]],
 [u'MACRO',
  [Identifier(u'scp')],
  [String(u'Science of Computer Programming')]],
 [u'MACRO', [Identifier(u'sicomp')], [String(u'SIAM Journal on Computing')]],
 [u'MACRO',
  [Identifier(u'tocs')],
  [String(u'ACM Transactions on Computer Systems')]],
 [u'MACRO',
  [Identifier(u'tods')],
  [String(u'ACM Transactions on Database Systems')]],
 [u'MACRO', [Identifier(u'tog')], [String(u'ACM Transactions on Graphics')]],
 [u'MACRO',
  [Identifier(u'toms')],
  [String(u'ACM Transactions on Mathematical Software')]],
 [u'MACRO',
  [Identifier(u'toois')],
  [String(u'ACM Transactions on Office Information Systems')]],
 [u'MACRO',
  [Identifier(u'toplas')],
  [String(u'ACM Transactions on Programming Languages and Systems')]],
 [u'MACRO', [Identifier(u'tcs')], [String(u'Theoretical Computer Science')]],
 [u'READ'],
 [u'FUNCTION',
  [Identifier(u'sortify')],
  [Identifier(u'purify$'), String(u'l'), Identifier(u'change.case$')]],
 [u'INTEGERS', [Identifier(u'len')]],
 [u'FUNCTION',
  [Identifier(u'chop.word')],
  [QuotedVar(u's'),
   Identifier(u':='),
   QuotedVar(u'len'),
   Identifier(u':='),
   Identifier(u's'),
   Integer(1),
   Identifier(u'len'),
   Identifier(u'substring$'),
   Identifier(u'='),
   FunctionLiteral([Identifier(u's'), Identifier(u'len'), Integer(1), Identifier(u'+'), Identifier(u'global.max$'), Identifier(u'substring$')]),
   QuotedVar(u's'),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'sort.format.names')],
  [QuotedVar(u's'),
   Identifier(u':='),
   Integer(1),
   QuotedVar(u'nameptr'),
   Identifier(u':='),
   String(u''),
   Identifier(u's'),
   Identifier(u'num.names$'),
   QuotedVar(u'numnames'),
   Identifier(u':='),
   Identifier(u'numnames'),
   QuotedVar(u'namesleft'),
   Identifier(u':='),
   FunctionLiteral([Identifier(u'namesleft'), Integer(0), Identifier(u'>')]),
   FunctionLiteral([Identifier(u'nameptr'), Integer(1), Identifier(u'>'), FunctionLiteral([String(u'   '), Identifier(u'*')]), QuotedVar(u'skip$'), Identifier(u'if$'), Identifier(u's'), Identifier(u'nameptr'), String(u'{vv{ } }{ll{ }}{  ff{ }}{  jj{ }}'), Identifier(u'format.name$'), QuotedVar(u't'), Identifier(u':='), Identifier(u'nameptr'), Identifier(u'numnames'), Identifier(u'='), Identifier(u't'), String(u'others'), Identifier(u'='), Identifier(u'and'), FunctionLiteral([String(u'et al'), Identifier(u'*')]), FunctionLiteral([Identifier(u't'), Identifier(u'sortify'), Identifier(u'*')]), Identifier(u'if$'), Identifier(u'nameptr'), Integer(1), Identifier(u'+'), QuotedVar(u'nameptr'), Identifier(u':='), Identifier(u'namesleft'), Integer(1), Identifier(u'-'), QuotedVar(u'namesleft'), Identifier(u':=')]),
   Identifier(u'while$')]],
 [u'FUNCTION',
  [Identifier(u'sort.format.title')],
  [QuotedVar(u't'),
   Identifier(u':='),
   String(u'A '),
   Integer(2),
   String(u'An '),
   Integer(3),
   String(u'The '),
   Integer(4),
   Identifier(u't'),
   Identifier(u'chop.word'),
   Identifier(u'chop.word'),
   Identifier(u'chop.word'),
   Identifier(u'sortify'),
   Integer(1),
   Identifier(u'global.max$'),
   Identifier(u'substring$')]],
 [u'FUNCTION',
  [Identifier(u'author.sort')],
  [Identifier(u'author'),
   Identifier(u'empty$'),
   FunctionLiteral([Identifier(u'key'), Identifier(u'empty$'), FunctionLiteral([String(u'to sort, need author or key in '), Identifier(u'cite$'), Identifier(u'*'), Identifier(u'warning$'), String(u'')]), FunctionLiteral([Identifier(u'key'), Identifier(u'sortify')]), Identifier(u'if$')]),
   FunctionLiteral([Identifier(u'author'), Identifier(u'sort.format.names')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'author.editor.sort')],
  [Identifier(u'author'),
   Identifier(u'empty$'),
   FunctionLiteral([Identifier(u'editor'), Identifier(u'empty$'), FunctionLiteral([Identifier(u'key'), Identifier(u'empty$'), FunctionLiteral([String(u'to sort, need author, editor, or key in '), Identifier(u'cite$'), Identifier(u'*'), Identifier(u'warning$'), String(u'')]), FunctionLiteral([Identifier(u'key'), Identifier(u'sortify')]), Identifier(u'if$')]), FunctionLiteral([Identifier(u'editor'), Identifier(u'sort.format.names')]), Identifier(u'if$')]),
   FunctionLiteral([Identifier(u'author'), Identifier(u'sort.format.names')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'author.organization.sort')],
  [Identifier(u'author'),
   Identifier(u'empty$'),
   FunctionLiteral([Identifier(u'organization'), Identifier(u'empty$'), FunctionLiteral([Identifier(u'key'), Identifier(u'empty$'), FunctionLiteral([String(u'to sort, need author, organization, or key in '), Identifier(u'cite$'), Identifier(u'*'), Identifier(u'warning$'), String(u'')]), FunctionLiteral([Identifier(u'key'), Identifier(u'sortify')]), Identifier(u'if$')]), FunctionLiteral([String(u'The '), Integer(4), Identifier(u'organization'), Identifier(u'chop.word'), Identifier(u'sortify')]), Identifier(u'if$')]),
   FunctionLiteral([Identifier(u'author'), Identifier(u'sort.format.names')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'editor.organization.sort')],
  [Identifier(u'editor'),
   Identifier(u'empty$'),
   FunctionLiteral([Identifier(u'organization'), Identifier(u'empty$'), FunctionLiteral([Identifier(u'key'), Identifier(u'empty$'), FunctionLiteral([String(u'to sort, need editor, organization, or key in '), Identifier(u'cite$'), Identifier(u'*'), Identifier(u'warning$'), String(u'')]), FunctionLiteral([Identifier(u'key'), Identifier(u'sortify')]), Identifier(u'if$')]), FunctionLiteral([String(u'The '), Integer(4), Identifier(u'organization'), Identifier(u'chop.word'), Identifier(u'sortify')]), Identifier(u'if$')]),
   FunctionLiteral([Identifier(u'editor'), Identifier(u'sort.format.names')]),
   Identifier(u'if$')]],
 [u'FUNCTION',
  [Identifier(u'presort')],
  [Identifier(u'type$'),
   String(u'book'),
   Identifier(u'='),
   Identifier(u'type$'),
   String(u'inbook'),
   Identifier(u'='),
   Identifier(u'or'),
   QuotedVar(u'author.editor.sort'),
   FunctionLiteral([Identifier(u'type$'), String(u'proceedings'), Identifier(u'='), QuotedVar(u'editor.organization.sort'), FunctionLiteral([Identifier(u'type$'), String(u'manual'), Identifier(u'='), QuotedVar(u'author.organization.sort'), QuotedVar(u'author.sort'), Identifier(u'if$')]), Identifier(u'if$')]),
   Identifier(u'if$'),
   String(u'    '),
   Identifier(u'*'),
   Identifier(u'year'),
   Identifier(u'field.or.null'),
   Identifier(u'sortify'),
   Identifier(u'*'),
   String(u'    '),
   Identifier(u'*'),
   Identifier(u'title'),
   Identifier(u'field.or.null'),
   Identifier(u'sort.format.title'),
   Identifier(u'*'),
   Integer(1),
   Identifier(u'entry.max$'),
   Identifier(u'substring$'),
   QuotedVar(u'sort.key$'),
   Identifier(u':=')]],
 [u'ITERATE', [Identifier(u'presort')]],
 [u'SORT'],
 [u'STRINGS', [Identifier(u'longest.label')]],
 [u'INTEGERS',
  [Identifier(u'number.label'), Identifier(u'longest.label.width')]],
 [u'FUNCTION',
  [Identifier(u'initialize.longest.label')],
  [String(u''),
   QuotedVar(u'longest.label'),
   Identifier(u':='),
   Integer(1),
   QuotedVar(u'number.label'),
   Identifier(u':='),
   Integer(0),
   QuotedVar(u'longest.label.width'),
   Identifier(u':=')]],
 [u'FUNCTION',
  [Identifier(u'longest.label.pass')],
  [Identifier(u'number.label'),
   Identifier(u'int.to.str$'),
   QuotedVar(u'label'),
   Identifier(u':='),
   Identifier(u'number.label'),
   Integer(1),
   Identifier(u'+'),
   QuotedVar(u'number.label'),
   Identifier(u':='),
   Identifier(u'label'),
   Identifier(u'width$'),
   Identifier(u'longest.label.width'),
   Identifier(u'>'),
   FunctionLiteral([Identifier(u'label'), QuotedVar(u'longest.label'), Identifier(u':='), Identifier(u'label'), Identifier(u'width$'), QuotedVar(u'longest.label.width'), Identifier(u':=')]),
   QuotedVar(u'skip$'),
   Identifier(u'if$')]],
 [u'EXECUTE', [Identifier(u'initialize.longest.label')]],
 [u'ITERATE', [Identifier(u'longest.label.pass')]],
 [u'FUNCTION',
  [Identifier(u'begin.bib')],
  [Identifier(u'preamble$'),
   Identifier(u'empty$'),
   QuotedVar(u'skip$'),
   FunctionLiteral([Identifier(u'preamble$'), Identifier(u'write$'), Identifier(u'newline$')]),
   Identifier(u'if$'),
   String(u'\\begin{thebibliography}{'),
   Identifier(u'longest.label'),
   Identifier(u'*'),
   String(u'}'),
   Identifier(u'*'),
   Identifier(u'write$'),
   Identifier(u'newline$')]],
 [u'EXECUTE', [Identifier(u'begin.bib')]],
 [u'EXECUTE', [Identifier(u'init.state.consts')]],
 [u'ITERATE', [Identifier(u'call.type$')]],
 [u'FUNCTION',
  [Identifier(u'end.bib')],
  [Identifier(u'newline$'),
   String(u'\\end{thebibliography}'),
   Identifier(u'write$'),
   Identifier(u'newline$')]],
 [u'EXECUTE', [Identifier(u'end.bib')]]]



#!/usr/bin/env python

# Copyright (c) 2006-2021  Andrey Golovizin
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from __future__ import unicode_literals

from pybtex.cmdline import CommandLine, make_option, standard_option


class PybtexConvertCommandLine(CommandLine):
    prog = 'pybtex-convert'
    args = '[options] in_filename out_filename'
    description = 'convert between bibliography database formats'
    long_description = """

pybtex-convert converts bibliography database files between supported formats
(currently BibTeX, BibTeXML and YAML).

    """.strip()

    num_args = 2

    options = (
        (None, (
            standard_option('strict'),
            make_option(
                '-f', '--from', dest='from_format',
                help='input format (%plugin_choices)', metavar='FORMAT',
                type='load_plugin', plugin_group='pybtex.database.input',
            ),
            make_option(
                '-t', '--to', dest='to_format',
                help='output format (%plugin_choices)', metavar='FORMAT',
                type='load_plugin', plugin_group='pybtex.database.output',
            ),
            standard_option('keyless_entries'),
            make_option(
                '--preserve-case', dest='preserve_case',
                action='store_true',
                help='do not convert identifiers to lower case',
            ),
        )),
        ('Encoding options', (
            standard_option('encoding'),
            standard_option('input_encoding'),
            standard_option('output_encoding'),
        )),
    )
    option_defaults = {
        'keyless_entries': False,
        'preserve_case': False,
    }

    def run(
        self, from_filename, to_filename,
        encoding, input_encoding, output_encoding,
        keyless_entries,
        **options
    ):
        from pybtex.database.convert import convert
        convert(
            from_filename, to_filename,
            input_encoding=input_encoding or encoding,
            output_encoding=output_encoding or encoding,
            parser_options={'keyless_entries': keyless_entries},
            **options
        )

main = PybtexConvertCommandLine()

if __name__ == '__main__':
    main()



# vim: fileencoding=utf-8
# Copyright (c) 2006-2021  Andrey Golovizin
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals
from __future__ import print_function
import re

try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

import six
import textwrap

from pybtex.exceptions import PybtexError
from pybtex.utils import (
    deprecated,
    OrderedCaseInsensitiveDict, CaseInsensitiveDefaultDict, CaseInsensitiveSet
)
from pybtex.richtext import Text
from pybtex.bibtex.utils import split_tex_string, scan_bibtex_string
from pybtex.errors import report_error
from pybtex.py3compat import fix_unicode_literals_in_doctest, python_2_unicode_compatible
from pybtex.plugin import find_plugin


# for python2 compatibility
def indent(text, prefix):
    if hasattr(textwrap, "indent"):
        return textwrap.indent(text, prefix)
    else:
        return ''.join(prefix + line for line in text.splitlines(True))


class BibliographyDataError(PybtexError):
    pass


class InvalidNameString(PybtexError):
    def __init__(self, name_string):
        message = 'Too many commas in {}'.format(repr(name_string))
        super(InvalidNameString, self).__init__(message)


class BibliographyData(object):
    def __init__(self, entries=None, preamble=None, wanted_entries=None, min_crossrefs=2):
        """
        A :py:class:`.BibliographyData` object contains a dictionary of bibliography
        entries referenced by their keys.
        Each entry represented by an :py:class:`.Entry` object.

        Additionally, :py:class:`.BibliographyData` may contain a LaTeX
        preamble defined by ``@PREAMBLE`` commands in the BibTeX file.
        """

        self.entries = OrderedCaseInsensitiveDict()
        '''A dictionary of bibliography entries referenced by their keys.

        The dictionary is case insensitive:

        >>> bib_data = parse_string("""
        ...     @ARTICLE{gnats,
        ...         author = {L[eslie] A. Aamport},
        ...         title = {The Gnats and Gnus Document Preparation System},
        ...     }
        ... """, 'bibtex')
        >>> bib_data.entries['gnats'] == bib_data.entries['GNATS']
        True

        '''

        self.crossref_count = CaseInsensitiveDefaultDict(int)
        self.min_crossrefs = min_crossrefs
        self._preamble = []
        if wanted_entries is not None:
            self.wanted_entries = CaseInsensitiveSet(wanted_entries)
            self.citations = CaseInsensitiveSet(wanted_entries)
        else:
            self.wanted_entries = None
            self.citations = CaseInsensitiveSet()
        if entries:
            if isinstance(entries, Mapping):
                entries = entries.items()
            for (key, entry) in entries:
                self.add_entry(key, entry)
        if preamble:
            self._preamble.extend(preamble)

    def __eq__(self, other):
        if not isinstance(other, BibliographyData):
            return super(BibliographyData, self) == other
        return (
            self.entries == other.entries
            and self._preamble == other._preamble
        )

    def __repr__(self):

        repr_entry = repr(self.entries)
        keys = self.entries.keys()

        for key in keys:
            ind = repr_entry.index(key) - 2  # find first instance
            repr_entry = repr_entry[:ind] + "\n" + repr_entry[ind:]

        repr_entry = indent(repr_entry, prefix="    ")
        repr_entry = repr_entry[4:]  # drop 1st indent

        return (
            "BibliographyData(\n"
            "  entries={0},\n\n"
            "  preamble={1})".format(repr_entry, repr(self._preamble))
        )

    def add_to_preamble(self, *values):
        self._preamble.extend(values)

    @property
    def preamble(self):
        r'''
        LaTeX preamble.

        >>> bib_data = parse_string(r"""
        ...     @PREAMBLE{"\newcommand{\noopsort}[1]{}"}
        ... """, 'bibtex')
        >>> print(bib_data.preamble)
        \newcommand{\noopsort}[1]{}

        .. versionadded:: 0.19
            Earlier versions used :py:meth:`.get_preamble()`, which is now deprecated.
        '''
        return ''.join(self._preamble)

    @deprecated('0.19', 'use BibliographyData.preamble instead')
    def get_preamble(self):
        """
        .. deprecated:: 0.19
            Use :py:attr:`.preamble` instead.
        """
        return self.preamble

    def want_entry(self, key):
        return (
            self.wanted_entries is None
            or key in self.wanted_entries
            or '*' in self.wanted_entries
        )

    def get_canonical_key(self, key):
        if key in self.citations:
            return self.citations.get_canonical_key(key)
        else:
            return key

    def add_entry(self, key, entry):
        if not self.want_entry(key):
            return
        if key in self.entries:
            report_error(BibliographyDataError('repeated bibliograhpy entry: %s' % key))
            return
        entry.key = self.get_canonical_key(key)
        self.entries[entry.key] = entry
        try:
            crossref = entry.fields['crossref']
        except KeyError:
            pass
        else:
            if self.wanted_entries is not None:
                self.wanted_entries.add(crossref)

    def add_entries(self, entries):
        for key, entry in entries:
            self.add_entry(key, entry)

    @fix_unicode_literals_in_doctest
    def _get_crossreferenced_citations(self, citations, min_crossrefs):
        r"""
        Get cititations not cited explicitly but referenced by other citations.

        >>> from pybtex.database import Entry
        >>> data = BibliographyData({
        ...     'main_article': Entry('article', {'crossref': 'xrefd_arcicle'}),
        ...     'xrefd_arcicle': Entry('article'),
        ... })
        >>> list(data._get_crossreferenced_citations([], min_crossrefs=1))
        []
        >>> list(data._get_crossreferenced_citations(['main_article'], min_crossrefs=1))
        [u'xrefd_arcicle']
        >>> list(data._get_crossreferenced_citations(['Main_article'], min_crossrefs=1))
        [u'xrefd_arcicle']
        >>> list(data._get_crossreferenced_citations(['main_article'], min_crossrefs=2))
        []
        >>> list(data._get_crossreferenced_citations(['xrefd_arcicle'], min_crossrefs=1))
        []

        >>> data2 = BibliographyData(data.entries, wanted_entries=data.entries.keys())
        >>> list(data2._get_crossreferenced_citations([], min_crossrefs=1))
        []
        >>> list(data2._get_crossreferenced_citations(['main_article'], min_crossrefs=1))
        [u'xrefd_arcicle']
        >>> list(data2._get_crossreferenced_citations(['Main_article'], min_crossrefs=1))
        [u'xrefd_arcicle']
        >>> list(data2._get_crossreferenced_citations(['main_article'], min_crossrefs=2))
        []
        >>> list(data2._get_crossreferenced_citations(['xrefd_arcicle'], min_crossrefs=1))
        []
        >>> list(data2._get_crossreferenced_citations(['xrefd_arcicle'], min_crossrefs=1))
        []

        """

        crossref_count = CaseInsensitiveDefaultDict(int)
        citation_set = CaseInsensitiveSet(citations)
        for citation in citations:
            try:
                entry = self.entries[citation]
                crossref = entry.fields['crossref']
            except KeyError:
                continue
            try:
                crossref_entry = self.entries[crossref]
            except KeyError:
                report_error(BibliographyDataError(
                    'bad cross-reference: entry "{key}" refers to '
                    'entry "{crossref}" which does not exist.'.format(
                        key=citation, crossref=crossref,
                    )
                ))
                continue

            canonical_crossref = crossref_entry.key
            crossref_count[canonical_crossref] += 1
            if crossref_count[canonical_crossref] >= min_crossrefs and canonical_crossref not in citation_set:
                citation_set.add(canonical_crossref)
                yield canonical_crossref

    @fix_unicode_literals_in_doctest
    def _expand_wildcard_citations(self, citations):
        r"""
        Expand wildcard citations (\citation{*} in .aux file).

        >>> from pybtex.database import Entry
        >>> data = BibliographyData((
        ...     ('uno', Entry('article')),
        ...     ('dos', Entry('article')),
        ...     ('tres', Entry('article')),
        ...     ('cuatro', Entry('article')),
        ... ))
        >>> list(data._expand_wildcard_citations([]))
        []
        >>> list(data._expand_wildcard_citations(['*']))
        [u'uno', u'dos', u'tres', u'cuatro']
        >>> list(data._expand_wildcard_citations(['uno', '*']))
        [u'uno', u'dos', u'tres', u'cuatro']
        >>> list(data._expand_wildcard_citations(['dos', '*']))
        [u'dos', u'uno', u'tres', u'cuatro']
        >>> list(data._expand_wildcard_citations(['*', 'uno']))
        [u'uno', u'dos', u'tres', u'cuatro']
        >>> list(data._expand_wildcard_citations(['*', 'DOS']))
        [u'uno', u'dos', u'tres', u'cuatro']

        """

        citation_set = CaseInsensitiveSet()
        for citation in citations:
            if citation == '*':
                for key in self.entries:
                    if key not in citation_set:
                        citation_set.add(key)
                        yield key
            else:
                if citation not in citation_set:
                    citation_set.add(citation)
                    yield citation

    def add_extra_citations(self, citations, min_crossrefs):
        expanded_citations = list(self._expand_wildcard_citations(citations))
        crossrefs = list(self._get_crossreferenced_citations(expanded_citations, min_crossrefs))
        return expanded_citations + crossrefs

    def to_string(self, bib_format, **kwargs):
        """
        Return the data as a unicode string in the given format.

        :param bib_format: Data format ("bibtex", "yaml", etc.).

        .. versionadded:: 0.19
        """
        writer = find_plugin('pybtex.database.output', bib_format)(**kwargs)
        return writer.to_string(self)

    @classmethod
    def from_string(cls, value, bib_format, **kwargs):
        """
        Return the data from a unicode string in the given format.

        :param bib_format: Data format ("bibtex", "yaml", etc.).

        .. versionadded:: 0.22.2
        """
        return parse_string(value, bib_format, **kwargs)

    def to_bytes(self, bib_format, **kwargs):
        """
        Return the data as a byte string in the given format.

        :param bib_format: Data format ("bibtex", "yaml", etc.).

        .. versionadded:: 0.19
        """
        writer = find_plugin('pybtex.database.output', bib_format)(**kwargs)
        return writer.to_bytes(self)

    def to_file(self, file, bib_format=None, **kwargs):
        """
        Save the data to a file.

        :param file: A file name or a file-like object.
        :param bib_format: Data format ("bibtex", "yaml", etc.).
            If not specified, Pybtex will try to guess by the file name.

        .. versionadded:: 0.19
        """
        if isinstance(file, six.string_types):
            filename = file
        else:
            filename = getattr(file, 'name', None)
        writer = find_plugin('pybtex.database.output', bib_format, filename=filename)(**kwargs)
        return writer.write_file(self, file)

    @fix_unicode_literals_in_doctest
    def lower(self):
        u'''
        Return another :py:class:`.BibliographyData` with all identifiers converted to lowercase.

        >>> data = parse_string("""
        ...     @BOOK{Obrazy,
        ...         title = "Obrazy z Rus",
        ...         author = "Karel Havlíček Borovský",
        ...     }
        ...     @BOOK{Elegie,
        ...         title = "Tirolské elegie",
        ...         author = "Karel Havlíček Borovský",
        ...     }
        ... """, 'bibtex')
        >>> data_lower = data.lower()
        >>> list(data_lower.entries.keys())
        [u'obrazy', u'elegie']
        >>> for entry in data_lower.entries.values():
        ...     entry.key
        ...     list(entry.persons.keys())
        ...     list(entry.fields.keys())
        u'obrazy'
        [u'author']
        [u'title']
        u'elegie'
        [u'author']
        [u'title']

        '''

        entries_lower = ((key.lower(), entry.lower()) for key, entry in self.entries.items())
        return type(self)(
            entries=entries_lower,
            preamble=self._preamble,
            wanted_entries=self.wanted_entries,
            min_crossrefs=self.min_crossrefs,
        )


class RichFieldProxyDict(Mapping):
    def __init__(self, fields):
        self._fields = fields

    def __contains__(self):
        return self._fields.__contains__()

    def __iter__(self):
        return self._fields.__iter__()

    def __len__(self):
        return self._fields.__len__()

    def __getitem__(self, key):
        return Text.from_latex(self._fields[key])


class Entry(object):
    """A bibliography entry."""

    type = None
    """Entry type (``'book'``, ``'article'``, etc.)."""

    key = None
    """Entry key (for example, ``'fukushima1980neocognitron'``)."""

    fields = None
    """A dictionary of entry fields.
    The dictionary is ordered and case-insensitive."""

    persons = None
    """
    A dictionary of entry persons, by their roles.

    The most often used roles are ``'author'`` and ``'editor'``.
    """

    """A reference to the containing :py:class:`.BibliographyData` object. Used to resolve crossrefs."""

    def __init__(self, type_, fields=None, persons=None):
        if fields is None:
            fields = {}
        if persons is None:
            persons = {}
        self.type = type_.lower()
        self.original_type = type_
        self.fields = OrderedCaseInsensitiveDict(fields)
        self.persons = OrderedCaseInsensitiveDict(persons)

    def __eq__(self, other):
        if not isinstance(other, Entry):
            return super(Entry, self) == other
        return (
            self.type == other.type
            and self.fields == other.fields
            and self.persons == other.persons
        )

    def __repr__(self):
        # represent the fields as a list of tuples for simplicity
        repr_fields = repr(list(self.fields.items()))
        keys = self.fields.keys()

        for key in keys:
            ind = repr_fields.index(key) - 2  # find first instance
            repr_fields = repr_fields[:ind] + "\n" + repr_fields[ind:]

        repr_fields = indent(repr_fields, prefix="    ")
        repr_fields = repr_fields[4:]  # drop 1st indent

        return (
            "Entry({0},\n"
            "  fields={1},\n"
            "  persons={2})".format(repr(self.type), repr_fields, repr(self.persons))
        )

    def add_person(self, person, role):
        self.persons.setdefault(role, []).append(person)

    def lower(self):
        return type(self)(
            self.type,
            fields=self.fields.lower(),
            persons=self.persons.lower(),
        )

    def _find_person_field(self, role):
        persons = self.persons[role]
        return ' and '.join(six.text_type(person) for person in persons)

    def _find_crossref_field(self, name, bib_data):
        if bib_data is None or 'crossref' not in self.fields:
            raise KeyError(name)
        referenced_entry = bib_data.entries[self.fields['crossref']]
        return referenced_entry._find_field(name, bib_data)

    def _find_field(self, name, bib_data=None):
        """
        Find the field with the given ``name`` according to this rules:

        - If the given field ``name`` in in ``self.fields``, just return
          self.fields[name].

        - Otherwise, if ``name`` is ``"authors"`` or ``"editors"`` (or any other
          person role), return the list of names as a string, separated by
          ``" and "``.

        - Otherwise, if this entry has a ``crossreff`` field, look up for the
          cross-referenced entry and try to find its field with the given
          ``name``.
        """
        try:
            return self.fields[name]
        except KeyError:
            try:
                return self._find_person_field(name)
            except KeyError:
                return self._find_crossref_field(name, bib_data)

    def to_string(self, bib_format, **kwargs):
        """
        Return the data as a unicode string in the given format.

        :param bib_format: Data format ("bibtex", "yaml", etc.).

        """
        writer = find_plugin('pybtex.database.output', bib_format)(**kwargs)
        return writer.to_string(BibliographyData(entries={self.key: self}))

    @classmethod
    def from_string(cls, value, bib_format, entry_number=0, **kwargs):
        """
        Return the data from a unicode string in the given format.

        :param bib_format: Data format ("bibtex", "yaml", etc.).
        :param entry_number: entry number if the string has more than one.

        .. versionadded:: 0.22.2
        """
        # get bibliography
        bibdata = BibliographyData.from_string(value, bib_format, **kwargs)
        # grab specific instance
        key = tuple(bibdata.entries.keys())[entry_number]
        return bibdata.entries[key]


@python_2_unicode_compatible
@fix_unicode_literals_in_doctest
class Person(object):
    """A person or some other person-like entity.

    >>> knuth = Person('Donald E. Knuth')
    >>> knuth.first_names
    [u'Donald']
    >>> knuth.middle_names
    [u'E.']
    >>> knuth.last_names
    [u'Knuth']

    """

    first_names = None
    """
    A list of first names.

    .. versionadded:: 0.19
        Earlier versions used :py:meth:`.first`, which is now deprecated.
    """

    middle_names = None
    """
    A list of middle names.

    .. versionadded:: 0.19
        Earlier versions used :py:meth:`.middle`, which is now deprecated.
    """

    prelast_names = None
    """
    A list of pre-last (aka von) name parts.

    .. versionadded:: 0.19
        Earlier versions used :py:meth:`.middle`, which is now deprecated.
    """

    last_names = None
    """
    A list of last names.

    .. versionadded:: 0.19
        Earlier versions used :py:meth:`.last`, which is now deprecated.
    """

    lineage_names = None
    """
    A list of linage (aka Jr) name parts.

    .. versionadded:: 0.19
        Earlier versions used :py:meth:`.lineage`, which is now deprecated.
    """

    valid_roles = ['author', 'editor']
    style1_re = re.compile(r'^(.+),\s*(.+)$')
    style2_re = re.compile(r'^(.+),\s*(.+),\s*(.+)$')

    def __init__(self, string="", first="", middle="", prelast="", last="", lineage=""):
        """
        :param string: The full name string.
            It will be parsed and split into separate first, last, middle,
            pre-last and lineage name parst.

            Supported name formats are:

            - von Last, First
            - von Last, Jr, First
            - First von Last

            (see BibTeX manual for explanation)

        """

        self.first_names = []
        self.middle_names = []
        self.prelast_names = []
        self.last_names = []
        self.lineage_names = []

        string = string.strip()
        if string:
            self._parse_string(string)
        self.first_names.extend(split_tex_string(first))
        self.middle_names.extend(split_tex_string(middle))
        self.prelast_names.extend(split_tex_string(prelast))
        self.last_names.extend(split_tex_string(last))
        self.lineage_names.extend(split_tex_string(lineage))

    @property
    @fix_unicode_literals_in_doctest
    def bibtex_first_names(self):
        """A list of first and middle names together.
        (BibTeX treats all middle names as first.)

        .. versionadded:: 0.19
            Earlier versions used :py:meth:`Person.bibtex_first`, which is now deprecated.


        >>> knuth = Person('Donald E. Knuth')
        >>> knuth.bibtex_first_names
        [u'Donald', u'E.']
        """
        return self.first_names + self.middle_names

    @fix_unicode_literals_in_doctest
    def _parse_string(self, name):
        """Extract various parts of the name from a string.

        >>> p = Person('Avinash K. Dixit')
        >>> print(p.first_names)
        [u'Avinash']
        >>> print(p.middle_names)
        [u'K.']
        >>> print(p.prelast_names)
        []
        >>> print(p.last_names)
        [u'Dixit']
        >>> print(p.lineage_names)
        []
        >>> print(six.text_type(p))
        Dixit, Avinash K.
        >>> p == Person(six.text_type(p))
        True
        >>> p = Person('Dixit, Jr, Avinash K. ')
        >>> print(p.first_names)
        [u'Avinash']
        >>> print(p.middle_names)
        [u'K.']
        >>> print(p.prelast_names)
        []
        >>> print(p.last_names)
        [u'Dixit']
        >>> print(p.lineage_names)
        [u'Jr']
        >>> print(six.text_type(p))
        Dixit, Jr, Avinash K.
        >>> p == Person(six.text_type(p))
        True

        >>> p = Person('abc')
        >>> print(p.first_names, p.middle_names, p.prelast_names, p.last_names, p.lineage_names)
        [] [] [] [u'abc'] []
        >>> p = Person('Viktorov, Michail~Markovitch')
        >>> print(p.first_names, p.middle_names, p.prelast_names, p.last_names, p.lineage_names)
        [u'Michail'] [u'Markovitch'] [] [u'Viktorov'] []
        """
        def process_first_middle(parts):
            try:
                self.first_names.append(parts[0])
                self.middle_names.extend(parts[1:])
            except IndexError:
                pass

        def process_von_last(parts):
            # von cannot be the last name in the list
            von_last = parts[:-1]
            definitely_not_von = parts[-1:]

            if von_last:
                von, last = rsplit_at(von_last, is_von_name)
                self.prelast_names.extend(von)
                self.last_names.extend(last)
            self.last_names.extend(definitely_not_von)

        def find_pos(lst, pred):
            for i, item in enumerate(lst):
                if pred(item):
                    return i
            return i + 1

        def split_at(lst, pred):
            """Split the given list into two parts.

            The second part starts with the first item for which the given
            predicate is True.
            """
            pos = find_pos(lst, pred)
            return lst[:pos], lst[pos:]

        def rsplit_at(lst, pred):
            rpos = find_pos(reversed(lst), pred)
            pos = len(lst) - rpos
            return lst[:pos], lst[pos:]

        def is_von_name(string):
            if string[0].isupper():
                return False
            if string[0].islower():
                return True
            else:
                for char, brace_level in scan_bibtex_string(string):
                    if brace_level == 0 and char.isalpha():
                        return char.islower()
                    elif brace_level == 1 and char.startswith('\\'):
                        return special_char_islower(char)
            return False

        def special_char_islower(special_char):
            control_sequence = True
            for char in special_char[1:]:  # skip the backslash
                if control_sequence:
                    if not char.isalpha():
                        control_sequence = False
                else:
                    if char.isalpha():
                        return char.islower()
            return False

        parts = split_tex_string(name, ',')
        if len(parts) > 3:
            report_error(InvalidNameString(name))
            last_parts = parts[2:]
            parts = parts[:2] + [' '.join(last_parts)]

        if len(parts) == 3:  # von Last, Jr, First
            process_von_last(split_tex_string(parts[0]))
            self.lineage_names.extend(split_tex_string(parts[1]))
            process_first_middle(split_tex_string(parts[2]))
        elif len(parts) == 2:  # von Last, First
            process_von_last(split_tex_string(parts[0]))
            process_first_middle(split_tex_string(parts[1]))
        elif len(parts) == 1:  # First von Last
            parts = split_tex_string(name)
            first_middle, von_last = split_at(parts, is_von_name)
            if not von_last and first_middle:
                last = first_middle.pop()
                von_last.append(last)
            process_first_middle(first_middle)
            process_von_last(von_last)
        else:
            # should hot really happen
            raise ValueError(name)

    def __eq__(self, other):
        if not isinstance(other, Person):
            return super(Person, self) == other
        return (
            self.first_names == other.first_names
            and self.middle_names == other.middle_names
            and self.prelast_names == other.prelast_names
            and self.last_names == other.last_names
            and self.lineage_names == other.lineage_names
        )

    def __str__(self):
        # von Last, Jr, First
        von_last = ' '.join(self.prelast_names + self.last_names)
        jr = ' '.join(self.lineage_names)
        first = ' '.join(self.first_names + self.middle_names)
        return ', '.join(part for part in (von_last, jr, first) if part)

    def __repr__(self):
        return 'Person({0})'.format(repr(six.text_type(self)))

    def get_part_as_text(self, type):
        names = getattr(self, type + '_names')
        return ' '.join(names)

    @fix_unicode_literals_in_doctest
    def get_part(self, type, abbr=False):
        """Get a list of name parts by `type`.

        >>> knuth = Person('Donald E. Knuth')
        >>> knuth.get_part('first')
        [u'Donald']
        >>> knuth.get_part('last')
        [u'Knuth']
        """

        names = getattr(self, type + '_names')
        if abbr:
            import warnings
            warnings.warn('Person.get_part(abbr=True) is deprecated since 0.19: use pybtex.textutils.abbreviate()', stacklevel=2)
            from pybtex.textutils import abbreviate
            names = [abbreviate(name) for name in names]
        return names

    @property
    def rich_first_names(self):
        """
        A list of first names converted to :ref:`rich text <rich-text>`.

        .. versionadded:: 0.20
        """

        return [Text.from_latex(name) for name in self.first_names]

    @property
    def rich_middle_names(self):
        """
        A list of middle names converted to :ref:`rich text <rich-text>`.

        .. versionadded:: 0.20
        """
        return [Text.from_latex(name) for name in self.middle_names]

    @property
    def rich_prelast_names(self):
        """
        A list of pre-last (aka von) name parts converted to :ref:`rich text <rich-text>`.

        .. versionadded:: 0.20
        """
        return [Text.from_latex(name) for name in self.prelast_names]

    @property
    def rich_last_names(self):
        """
        A list of last names converted to :ref:`rich text <rich-text>`.

        .. versionadded:: 0.20
        """
        return [Text.from_latex(name) for name in self.last_names]

    @property
    def rich_lineage_names(self):
        """
        A list of lineage (aka Jr) name parts converted to :ref:`rich text <rich-text>`.

        .. versionadded:: 0.20
        """
        return [Text.from_latex(name) for name in self.lineage_names]

    @deprecated('0.19', 'use Person.first_names instead')
    def first(self, abbr=False):
        """
        .. deprecated:: 0.19
            Use :py:attr:`.first_names` instead.
        """
        return self.get_part('first', abbr)

    @deprecated('0.19', 'use Person.middle_names instead')
    def middle(self, abbr=False):
        """
        .. deprecated:: 0.19
            Use :py:attr:`.middle_names` instead.
        """
        return self.get_part('middle', abbr)

    @deprecated('0.19', 'use Person.prelast_names instead')
    def prelast(self, abbr=False):
        """
        .. deprecated:: 0.19
            Use :py:attr:`.prelast_names` instead.
        """
        return self.get_part('prelast', abbr)

    @deprecated('0.19', 'use Person.last_names instead')
    def last(self, abbr=False):
        """
        .. deprecated:: 0.19
            Use :py:attr:`.last_names` instead.
        """
        return self.get_part('last', abbr)

    @deprecated('0.19', 'use Person.lineage_names instead')
    def lineage(self, abbr=False):
        """
        .. deprecated:: 0.19
            Use :py:attr:`.lineage_names` instead.
        """
        return self.get_part('lineage', abbr)

    @deprecated('0.19', 'use Person.bibtex_first_names instead')
    def bibtex_first(self):
        """
        .. deprecated:: 0.19
            Use :py:attr:`.bibtex_first_names` instead.
        """
        return self.bibtex_first_names


def parse_file(file, bib_format=None, **kwargs):
    """
    Read bibliography data from file and return a :py:class:`.BibliographyData` object.

    :param file: A file name or a file-like object.
    :param bib_format: Data format ("bibtex", "yaml", etc.).
        If not specified, Pybtex will try to guess by the file name.

    .. versionadded:: 0.19
    """

    if isinstance(file, six.string_types):
        filename = file
    else:
        filename = getattr(file, 'name', None)

    parser = find_plugin('pybtex.database.input', bib_format, filename=filename)(**kwargs)
    return parser.parse_file(file)


def parse_string(value, bib_format, **kwargs):
    """
    Parse a Unicode string containing bibliography data and return a :py:class:`.BibliographyData` object.

    :param value: Unicode string.
    :param bib_format: Data format ("bibtex", "yaml", etc.).

    .. versionadded:: 0.19
    """

    parser = find_plugin('pybtex.database.input', bib_format)(**kwargs)
    return parser.parse_string(value)


def parse_bytes(value, bib_format, **kwargs):
    """
    Parse a byte string containing bibliography data and return a :py:class:`.BibliographyData` object.

    :param value: Byte string.
    :param bib_format: Data format (for example, "bibtexml").

    .. versionadded:: 0.19
    """

    parser = find_plugin('pybtex.database.input', bib_format)(**kwargs)
    return parser.parse_bytes(value)



#!/usr/bin/env python

# Copyright (c) 2006-2021  Andrey Golovizin
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals

from os import path

from pybtex.cmdline import CommandLine, make_option, standard_option


class PybtexCommandLine(CommandLine):
    prog = 'pybtex'
    args = '[options] auxfile.aux'
    description = 'BibTeX-compatible bibliography processor in Python'
    long_description = """

Pybtex reads citation information from a LaTeX .aux file and produces a
formatted bibliography. Pybtex understands BibTeX .bib and .bst style files and
can be used as a drop-in replacement for BibTeX.

Besides BibTeX .bib files, BibTeXML and YAML bibliography files are
supported.

It is also possible to define bibliography formatting styles in Python.

    """.strip()
    num_args = 1

    options = (
        (None, (
            standard_option('strict'),
            make_option(
                '--terse', dest='verbose', action='store_false',
                help='ignored for compatibility with BibTeX',
            ),
            standard_option('min_crossrefs'),
            standard_option('bib_format'),
            standard_option('output_backend'),
            standard_option('style'),
            make_option(
                '-l', '--style-language', dest='style_language',
                help='style definition language to use (bibtex or python)',
                metavar='LANGUAGE',
            ),
        )),
        ('Pythonic style options', (
            standard_option('label_style'),
            standard_option('name_style'),
            standard_option('sorting_style'),
            standard_option('abbreviate_names'),
        )),
        ('Encoding options', (
            standard_option('encoding'),
            make_option('--bibtex-encoding', dest='bib_encoding', metavar='ENCODING'),
            make_option('--bst-encoding', dest='bst_encoding', metavar='ENCODING'),
            standard_option('output_encoding'),
        )),
    )
    option_defaults = {
        'style_language': 'bibtex',
        'min_crossrefs': 2,
    }
    legacy_options = '-help', '-version', '-min-crossrefs', '-terse'

    def run(self, filename, style_language, encoding, **options):
        if style_language == 'bibtex':
            from pybtex import bibtex as engine
        elif style_language == 'python':
            import pybtex as engine
        else:
            self.opt_parser.error('unknown style language %s' % style_language)

        not_supported_by_bibtex = {
            'output_backend': 'output backends',
            'label_style': 'label styles',
            'name_style': 'name styles',
            'sorting_style': 'sorting styles',
            'abbreviate_names': 'abbreviated names',
        }
        if style_language != 'python':
            for option, what_is_not_supported in not_supported_by_bibtex.items():
                if options[option]:
                    self.opt_parser.error(
                        '%s are only supported by the Pythonic style engine (-l python)' % what_is_not_supported
                    )

        for encoding_option in 'bib_encoding', 'bst_encoding', 'output_encoding':
            if not options[encoding_option]:
                options[encoding_option] = encoding

        ext = path.splitext(filename)[1]
        if ext != '.aux':
            filename = path.extsep.join([filename, 'aux'])
        engine.make_bibliography(filename, **options)

main = PybtexCommandLine()

if __name__ == '__main__':
    main()
